unit View3code;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, MPlayer;

type
  TMain = class(TForm)
    ExitBtn: TBitBtn;
    StartBtn: TButton;
    Timer: TTimer;
    Image: TImage;
    ZoomBtn: TButton;
    TripBtn: TButton;
    MediaPlayer: TMediaPlayer;
    Button: TButton;
    ShapeBtn: TButton;

//***********************************************
//**** Godamn Procs
//***********************************************
    {creationism}
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure CreateUniverse;
    procedure FillMathsArrays;
    procedure FillUpData;

    {Button Controls}
    procedure ExitBtnClick(Sender: TObject);
    procedure ShapeBtnClick(Sender: TObject);
    procedure StartBtnClick(Sender: TObject);
    procedure TripBtnClick(Sender: TObject);
    procedure ZoomBtnClick(Sender: TObject);

    {Main Body}
    procedure CalcDrawOrder;
    procedure CalcMatrix;
    procedure CalcNormals;
    procedure DrawPoly;
    procedure FillPolyPoints;
    procedure Transform;
    procedure TimerTimer(Sender: TObject);

    {Key Controls}
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure AdjustAngle1(amount: Integer);
    procedure AdjustAngle2(amount: Integer);
    procedure AdjustAngle3(amount: Integer);
    procedure MediaPlayerNotify(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);

  private
    { Private declarations }
  public
    { Public declarations }
  end;

//***********************************************
//**** Constants
//***********************************************
const Accuracy       = 720;
      CNumPoints     = 100;
      CNumEdges      = 200;
      CNumPolyPoints = 20;
      CNumPolys      = 100;
//***********************************************
//**** Variables
//***********************************************
var
  Main: TMain;
  MainPath: String;
  DataPath: String;

  //double buffer
  ARect : TRect;
  BufferShowing : Integer;
  Buffer : TBitmap;
  Bitmap1 : TBitmap;
  Bitmap2 : TBitmap;
  Doubling : Boolean;
  OldTime : String;
  FrameCount : Integer;

  // Screen Polygon Points
  PolyPoints: Array [0..CNumPolys, 0..CNumPolyPoints] of TPoint;

  //maths array
  CosArray: Array [0..Accuracy] of Integer;
  SinArray: Array [0..Accuracy] of Integer;

  // Point Arrays
  PX: Array [0..CNumPoints] of Integer;
  PY: Array [0..CNumPoints] of Integer;
  PZ: Array [0..CNumPoints] of Integer;

  NX: Array [0..CNumPolys] of Extended;
  NY: Array [0..CNumPolys] of Extended;
  NZ: Array [0..CNumPolys] of Extended;

  // Edge List Array
  E: Array [0..CNumEdges] of Integer;

  // Number of Edges Array
  NumEdges: Array [0..CNumPolys] of Integer;

  // Edge Offset Array
  EdgeOffset: Array [0..CNumPolys] of Integer;
  Offset : Integer;

  // Colour Array
  PolyColour: Array [0..CNumPolys] of Integer;
  IlumArray: Array [0..CNumPolys] of Extended;
  AmbientLight: Byte;

  // Globals
  FarCoord: Array [0..CNumPolys] of Integer;
  DrawOrder: Array [0..CNumPolys] of Integer;
  DrawOrderTemp: Array [0..5,0..CNumPolys] of Integer;
  ShapeFile: Array [0..10] of String;
  ShapeCount: Integer;

  NumPolys: Integer;
  DrawingPolyNoEdges : Integer;
  DrawingPoly : Integer;

  CenterX : Integer;
  CenterY : Integer;
  Colour : Integer;
  DrawIt : Boolean;

  Zoom : Integer;
  Zooming : Boolean;
  ZoomDir : Integer;
  ZoomSpeed : Integer;

  // Angles
  Angle1 : Integer;
  Angle2 : Integer;
  Angle3 : Integer;
  Angle1Speed : Integer;
  Angle2Speed : Integer;
  Angle3Speed : Integer;
  spin : Boolean;
  
  // Matrix Maths
  s1, s2, s3 : Integer;
  c1, c2, c3 : Integer;
  scale      : Integer;

  x2 : Integer;
  y1 : Integer;
  z1 : Integer;

  tx, ty, tz : Integer;

  x,y,z      : Integer;
  ex,ey,ez   : Extended;
  ex1,ey1,ez1   : Extended;
  scx, scy : Extended;

  denom :  Integer;
  persp : Integer;
  ilum : Extended;
  ptr : ^byte;

  {normals}
  lsx, lsy, lsz: Integer;
  nsx, nsy, nsz: Extended;
  NormU : Extended;
  lx, ly, lz : Extended;
  LOSU : Extended;

  offx : Integer;
  offy : Integer;
  num : Integer;
  counter : Integer;
  placed : Boolean;
  
  px1, py1, px2, py2 : Integer;
implementation

{$R *.DFM}

//***********************************************
//**** Creationism Rules OK
//***********************************************
procedure TMain.FormCreate(Sender: TObject);
begin
  MainPath := ExtractFilePath(Application.ExeName);
  DataPath := MainPath+'Data';

  {Set Screen and Control Defaults}
  Left := 0;
  Top := 0;
  Width := Screen.Width;
  Height := Screen.Height;
  Image.Left := 0;
  Image.Top := 0;
  Image.Width := Width;
  Image.Height := Height;

  ExitBtn.Left := 2;
  ExitBtn.Top := Height - 27;
  TripBtn.Left := 2;
  TripBtn.Top := ExitBtn.Top - 28;
  ShapeBtn.Left := 2;
  ShapeBtn.Top := TripBtn.Top - 28;
  ZoomBtn.Left := 2;
  ZoomBtn.Top := ShapeBtn.Top - 28;
  StartBtn.Left := 2;
  StartBtn.Top := ZoomBtn.Top - 28;

  {create doublebuffer}
  BufferShowing := 1;
  Bitmap1 := TBitmap.Create;
  Bitmap2 := TBitmap.Create;
  Bitmap1.Width := Width;
  Bitmap1.Height := Height;
  Bitmap2.Width := Width;
  Bitmap2.Height := Height;
  Doubling := True;
  FrameCount := 0;
  OldTime := '00:00:00';

  Bitmap1.Canvas.CopyMode := cmBlackness;
  Bitmap2.Canvas.CopyMode := cmBlackness;
  ARect := Rect(0, 0, Bitmap1.Width, Bitmap1.Height);	{ get bitmap rectangle for blanking buffer}

  Image.Picture.Graphic := Bitmap1;

  {Set Globals}
  ShapeFile[0] := 'cube.txt';
  ShapeFile[1] := 'jake.txt';
  ShapeFile[2] := 'pyramid.txt';
  ShapeFile[3] := 'ship.txt';
  ShapeFile[4] := 'cube2.txt';
  ShapeFile[5] := 'oct.txt';
  ShapeFile[6] := 'enter.txt';
  ShapeCount   := 0;

  AmbientLight := byte($40);
  CenterX := Width div 2;
  CenterY := Height div 2;
  Zoom := -300;
  Zooming := False;
  ZoomDir := 0;
  ZoomSpeed := 50;

  {Set AngleStuff}
  Angle1 := 0;
  Angle2 := 0;
  Angle3 := 0;
  Angle1Speed := 8;
  Angle2Speed := 6;
  Angle3Speed := 4;
  spin := False;
  
  {Set Maths}
  scale := 65536;
  persp := 700;

  offx := 100;
  offy := 100;
  num := 0;
end;

//***************************
procedure TMain.FormActivate(Sender: TObject);
begin
  {setup drawing defaults}
  FillMathsArrays;
  CreateUniverse;
end;

//***************************
procedure TMain.CreateUniverse;
begin
  FillUpData;
  CalcNormals;
  CalcDrawOrder;

  StartBtn.SetFocus;
end;

//***************************
procedure TMain.FillMathsArrays;
var n: Integer;
begin
  For n := 0 to Accuracy do
  begin
    CosArray[n] := round(Cos((n/accuracy)*(2*pi))*scale);
    SinArray[n] := round(Sin((n/accuracy)*(2*pi))*scale);
  end;
end;

//***************************
procedure TMain.FillUpData;
var n      : Integer;
    PolyNo : Integer;
    NoEdges: Integer;
    F      : TextFile;
    S      : String;
    maxcol : Byte;
begin
  AssignFile(F, ShapeFile[ShapeCount]);
  Reset(F);

  {read shape points into array}
  n := 0;
  ReadLn(F, S);

  While not (S='end') do
  begin
    PX[n] := StrToInt(S);
    ReadLn(F, S);
    PY[n] := StrToInt(S);
    ReadLn(F, S);
    PZ[n] := StrToInt(S);
    ReadLn(F, S);
    Inc(n);
  end;

  {read edge list into array}
  PolyNo  := 0;
  NoEdges := 0;
  n       := 0;
  EdgeOffset[0] := 0;
  ReadLn(F, S);

  While not (S='end') do
  begin
    If S <> '-1' then
      E[n] := StrToInt(S)
    else
    begin
      NumEdges[PolyNo] := NoEdges-1;
      NoEdges := -1;
      ReadLn(F, S);
      PolyColour[PolyNo] := StrToInt(S);
      maxcol := byte($ff) - AmbientLight;
      Case PolyColour[PolyNo] of
        0 : PolyColour[PolyNo] := $00000000;
        1 : PolyColour[PolyNo] := 0 + maxcol;
        2 : PolyColour[PolyNo] := 0 + maxcol*256;
        3 : PolyColour[PolyNo] := 0 + maxcol*65536;
        4 : PolyColour[PolyNo] := 0 + maxcol*256 + maxcol;
        5 : PolyColour[PolyNo] := 0 + maxcol*65536 + maxcol*256;
        6 : PolyColour[PolyNo] := 0 + maxcol*65536 + maxcol;
        7 : PolyColour[PolyNo] := 0 + maxcol*65536 + maxcol*256 + maxcol;
      end;
      Inc(PolyNo);
      EdgeOffset[PolyNo] := n+1;
    end;

    Inc(n);
    Inc(NoEdges);
    ReadLn(F, S);
  end;

  NumPolys := PolyNo-1;

  CloseFile(F);
  exit;
end;

//***************************
//**** Button Controls
//***************************
procedure TMain.ExitBtnClick(Sender: TObject);
begin
  Bitmap1.Free;
  Bitmap2.Free;
  Close;
end;

//***************************
procedure TMain.ShapeBtnClick(Sender: TObject);
begin
    Timer.Enabled := False;
    Inc(ShapeCount);
    If ShapeCount = 7 then
      ShapeCount := 0;
    CreateUniverse;
    Timer.Enabled := True;
end;
//***************************

procedure TMain.StartBtnClick(Sender: TObject);
begin
  MediaPlayer.FileName := MainPath + 'Mercyful.wav';
  MediaPlayer.Open;
  MediaPlayer.Notify := True;
  MediaPlayer.Play;

  Timer.Enabled := True;
end;

//***************************
procedure TMain.TripBtnClick(Sender: TObject);
begin
  Doubling := not(Doubling);
end;

//***************************
procedure TMain.ZoomBtnClick(Sender: TObject);
begin
  Zooming := not(Zooming);
end;

//***************************
//**** Main Body
//***************************
procedure TMain.CalcDrawOrder;
var n, view, j ,k, li, total : Integer;
    Offset: Integer;
    num: array [0..6] of Integer;
    Replace, Found: Boolean;
    DrawCounter: Integer;
begin
  For n := 0 to NumPolys do
  begin
//    DrawOrder[n] := NumPolys;
    DrawOrder[n] := n;
    For view := 0 to 5 do
    begin
      DrawOrderTemp[view,n] := NumPolys;
      num[view] := 0;
    end;
  end;

(*  For view := 0 to 5 do
  begin

  For n := 0 to NumPolys do
  begin
    FarCoord[n] := 10000;
  end;

  For n := 0 to NumPolys do
  begin
    {get polygon's furthest z coord}
    For j := 0 to NumEdges[n] do
    begin
      Offset := E[EdgeOffset[n]+j];
      Case View of
      0 : z := Pz[Offset];
      1 : z := -Pz[Offset];
      2 : z := Px[Offset];
      3 : z := -Px[Offset];
      4 : z := Py[Offset];
      5 : z := -Py[Offset];
      end;
      If z < FarCoord[n] then
        FarCoord[n] := z;
    end;

    For k := 0 to NumPolys do
      If FarCoord[n] < FarCoord[DrawOrderTemp[view,k]] then
      begin
        For li := NumPolys-1 DownTo k do
          DrawOrderTemp[view,li+1] := DrawOrderTemp[view,li];
        Replace := False;
        Case View of
        0 : If Nz[n] >= 0 then
            Replace := True;
        1 : If Nz[n] <= 0 then
            Replace := True;
        2 : If Nx[n] >= 0 then
            Replace := True;
        3 : If Nx[n] <= 0 then
            Replace := True;
        4 : If Ny[n] >= 0 then
            Replace := True;
        5 : If Ny[n] <= 0 then
            Replace := True;
        end;
        If replace then
        begin
          DrawOrderTemp[view,k] := n;
          inc(num[view]);
        end;
        break;
      end;
  end; // end of NumPolys loop
  end; // end of view loop

  DrawCounter := 0;
  For view := 0 to 5 do
  begin
    If num[view] > 0 then
    For j := 0 to num[view]-1 do
    begin
      //see if poly j is already in DrawOrder
      Found := False;
      For n := 0 to NumPolys do
        If DrawOrder[n] = DrawOrderTemp[view,j] then
          Found := True;

      //if not yet placed then add poly to Draw Order list
      If not Found then
      begin
        DrawOrder[DrawCounter] := DrawOrderTemp[view,j];
        inc(DrawCounter);
      end;

      //if all polys placed then stop this charade
      If DrawCounter = NumPolys then
        break;
    end;
  end; //end of view loop
 *)
end;
//***************************

procedure TMain.CalcMatrix;
begin
  s1 := SinArray[angle1];
  s2 := SinArray[angle2];
  s3 := SinArray[angle3];
  c1 := CosArray[angle1];
  c2 := CosArray[angle2];
  c3 := CosArray[angle3];
end;

//***************************
procedure TMain.CalcNormals;
var
  n : Integer;
  l1x, l1y, l1z: Integer;  // line 1 vector
  l2x, l2y, l2z: Integer;  // line 2 vector
  nlx, nly, nlz: Extended; // normal line vector
begin
  For n := 0 to NumPolys do
  begin
    Offset := EdgeOffset[n];

    l1x := PX[E[Offset+1]] - PX[E[Offset+0]];
    l1y := PY[E[Offset+1]] - PY[E[Offset+0]];
    l1z := PZ[E[Offset+1]] - PZ[E[Offset+0]];
    l2x := PX[E[Offset+2]] - PX[E[Offset+1]];
    l2y := PY[E[Offset+2]] - PY[E[Offset+1]];
    l2z := PZ[E[Offset+2]] - PZ[E[Offset+1]];

    {get normal vector - too large}
    nlx := l1y*l2z - l1z*l2y;
    nly := l1z*l2x - l1x*l2z;
    nlz := l1x*l2y - l1y*l2x;

    {get Normal Unit vector}
    NormU := Sqrt(nlx*nlx + nly*nly + nlz*nlz);
    If NormU = 0 then NormU := 0.0001;
    NX[n] := nlx/NormU;
    NY[n] := nly/NormU;
    NZ[n] := nlz/NormU;
  end;
end;

//***************************
procedure TMain.DrawPoly;
var n,i,j, startnum : Integer;
begin
  {swap buffers}
  If BufferShowing = 1 then
  begin
    BufferShowing := 2;
    Image.Picture.Graphic := Bitmap1;
    Buffer := Bitmap2;
  end
  else
  begin
    BufferShowing := 1;
    Image.Picture.Graphic := Bitmap2;
    Buffer := Bitmap1;
  end;

  {clear buffer}
  If Doubling then
    Buffer.Canvas.CopyRect(ARect, Buffer.Canvas, ARect);	{ copy bitmap over itself }

  CalcMatrix;

  startnum := -(num div 2);
  For i := startnum  to startnum + num do
  begin
    offx := i*280;
  if num mod 2 <> 0 then
    offx := offx - 175;
  For j := startnum to startnum + num do
  begin
    offy := j*280;
  if num mod 2 <> 0 then
    offy := offy - 175;

  {calc all visible polys coords}
  Counter := 0;
  For n := 0 to NumPolys do
  begin
    DrawingPoly := n;
    DrawingPolyNoEdges := NumEdges[DrawingPoly];
    FillPolyPoints;
  end;

  {draw polys in correct z-order}
  For n := 0 to Counter-1 do
  begin
    DrawingPoly := DrawOrder[n];
    DrawingPolyNoEdges := NumEdges[DrawingPoly];

    With Buffer.Canvas Do
    begin
      {set polygon colour}
      ilum := IlumArray[DrawingPoly];
      If ilum < 0  then ilum := 0;

      Colour := PolyColour[DrawingPoly];
      ptr := @Colour;
      If ptr^ > 0 then
      begin
        ptr^ := round(ilum*ptr^);
        If ptr^ < AmbientLight then
           ptr^ := byte(AmbientLight);
      end;
      inc(ptr);
      If ptr^ > 0 then
      begin
        ptr^ := round(ilum*ptr^);
        If ptr^ < AmbientLight then
           ptr^ := byte(AmbientLight);
      end;
      inc(ptr);
      If ptr^ > 0 then
      begin
        ptr^ := round(ilum*ptr^);
        If ptr^ < AmbientLight then
           ptr^ := byte(AmbientLight);
      end;
      Brush.Color := colour;
      Pen.Color := clBlack;

      Polygon(Slice(PolyPoints[DrawingPoly], DrawingPolyNoEdges+1));
    end;
  end; //end of main for loop

  end;
  end;
end;

//***************************
procedure TMain.FillPolyPoints;
var n, k, li: Integer;
    farz: Integer;
begin
  farz := -10000;
  {calc screen coords}
  For n := 0 to DrawingPolyNoEdges do
  begin
    Offset := E[EdgeOffset[DrawingPoly]+n];

    // starting point
    x := PX[Offset];
    y := PY[Offset];
    z := PZ[Offset];
    Transform;
    // calc perspective ratio
    denom := tz + persp;
    If denom = 0 then
     denom := 1;
    // is it farthest z ?
    If tz>farz then
      farz := tz;

    // plot screen point with perspective
    PolyPoints[DrawingPoly, n].X := CenterX + (tx * persp) div denom;
    PolyPoints[DrawingPoly, n].Y := CenterY + (ty * persp) div denom;

    //check to see if face is visible or not}
    If n = 2 then
    begin
      px1 := PolyPoints[DrawingPoly,1].X - PolyPoints[DrawingPoly,0].X;
      py1 := PolyPoints[DrawingPoly,1].Y - PolyPoints[DrawingPoly,0].Y;
      px2 := PolyPoints[DrawingPoly,2].X - PolyPoints[DrawingPoly,1].X;
      py2 := PolyPoints[DrawingPoly,2].Y - PolyPoints[DrawingPoly,1].Y;

      If (px2*py1 - py2*px1) > 0 then
        DrawIt := True
      else
      begin
        DrawIt := False;
        break;
      end;
    end;
  end; // end of calc screen coords

  {z buffering}
  If DrawIt = True then
  begin
    FarCoord[DrawingPoly] := farz;
    If counter = 0 then
      DrawOrder[0] := DrawingPoly
    else
    begin
      //put this poly in the right place in the queue
      placed := False;
      For k := 0 to counter-1 do
      If farz > FarCoord[DrawOrder[k]] then
      begin
        For li := NumPolys-1 DownTo k do
          DrawOrder[li+1] := DrawOrder[li];
        DrawOrder[k] := DrawingPoly;
        Placed := True;
        break;
      end;
      If not placed then
        DrawOrder[counter] := DrawingPoly;
    end;
    inc(counter);
  end
  else
    exit;

  {rotate Normal point}
  ex := NX[DrawingPoly];
  ey := NY[DrawingPoly];
  ez := NZ[DrawingPoly];
  //// x axis ////
  ey1 := (ey*c1 + ez*-s1) / scale;
  ez1 := (ey*s1 + ez*c1) / scale;
  //// y axis ////
  ex1 := (ex*c2 + ez1*s2) / scale;
  nsz := (ex*-s2 + ez1*c2) / scale;
  //// z axis ////
  nsx := (ex1*c3 + ey1*-s3) / scale;
  nsy := (ex1*s3 + ey1*c3) / scale;

  {calc Line Of Sight point ... centre point (0,0,0) of polygon}
  tz := -Zoom - persp;
  tx := -offx;
  ty := -offy;

  {get Line Of Sight Unit vector}
  LOSU := Sqrt(tx*tx + ty*ty + tz*tz);
  If LOSU = 0 then LOSU := 0.0001;
  lx := tx/LOSU;
  ly := ty/LOSU;
  lz := tz/LOSU;

  {calc illumination}
  IlumArray[DrawingPoly] := lx*nsx + ly*nsy + lz*nsz;
end;

//***************************
procedure TMain.Transform;
begin
    //// x axis ////
    y1 := (y*c1 + z*-s1) div scale;
    z1 := (y*s1 + z*c1) div scale;
    //// y axis ////
    x2 := (x*c2 + z1*s2) div scale;
    tz := (x*-s2 + z1*c2) div scale;
    //// z axis ////
    tx := (x2*c3 + y1*-s3) div scale;
    ty := (x2*s3 + y1*c3) div scale;

    ////movement
    tz := tz + Zoom;
    tx := tx + offx;
    ty := ty + offy;
end;

//***************************
procedure TMain.TimerTimer(Sender: TObject);
begin
  inc(FrameCount);
  If Copy(TimeToStr(time), 1, 8)>oldtime then
  begin
    oldtime := Copy(TimeToStr(time), 1, 8);
    Button.Caption := IntToStr(FrameCount);
    FrameCount := 0;
  end;

 AdjustAngle1(Angle1Speed);
 AdjustAngle2(Angle2Speed);
 AdjustAngle3(Angle3Speed);

  If Zooming then
  begin
    If ZoomDir = 1 then
      Zoom := Zoom + ZoomSpeed
    else
      Zoom := Zoom - ZoomSpeed;
    If Zoom > 4000 then
      ZoomDir := 0;
    If Zoom < 200 then
      ZoomDir := 1;
  end;

  DrawPoly;
end;

//***********************************************
//**** Key Controls
//***********************************************
procedure TMain.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  If (ssCtrl in Shift) and ((Key=ord('x')) or (Key=ord('X'))) then
    ExitBtnClick(ExitBtn);
  If (Key>=49) and (Key<=57) then
    num := Key-49;
  If (Key = ord('t')) or (Key = ord('T')) then
    TripBtnClick(TripBtn);
  If (Key = ord('z')) or (Key = ord('Z')) then
    ZoomBtnClick(ZoomBtn);
  If (Key = ord('f')) or (Key = ord('F')) then
    ShapeBtnClick(ShapeBtn);
end;
//***************************
procedure TMain.FormKeyPress(Sender: TObject; var Key: Char);
begin
  If Key = 'j' then
     AdjustAngle1(-8);
  If Key = 'n' then
     AdjustAngle1(8);
  If Key = 'm' then
     AdjustAngle2(-8);
  If Key = 'b' then
     AdjustAngle2(8);
  If Key = 'k' then
     AdjustAngle3(-8);
  If Key = 'h' then
     AdjustAngle3(8);
  If Key = 's' then
     Spin := not Spin;
end;

procedure TMain.AdjustAngle1(amount: Integer);
begin
  Angle1 := Angle1 + amount;
  If Angle1 >= Accuracy then
    Dec(Angle1, Accuracy);
  If Angle1 <= 0 then
    Inc(Angle1, Accuracy);
end;

procedure TMain.AdjustAngle2(amount: Integer);
begin
  Angle2 := Angle2 + amount;
  If Angle2 >= Accuracy then
    Dec(Angle2, Accuracy);
  If Angle2 < 0 then
    Inc(Angle2, Accuracy);
end;

procedure TMain.AdjustAngle3(amount: Integer);
begin
  Angle3 := Angle3 + amount;
  If Angle3 >= Accuracy then
    Dec(Angle3, Accuracy);
  If Angle3 < 0 then
    Inc(Angle3, Accuracy);
end;

procedure TMain.MediaPlayerNotify(Sender: TObject);
begin
  MediaPlayer.Notify := True;
  MediaPlayer.Play;
end;


end.
